from enum import Enum
from typing import Annotated, Any, cast, Generic, List, Literal, Optional, TypedDict, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr
from pydantic.generics import GenericModel

from signals_notebook.common_types import DateTime, EID, EntityType, MaterialType, MID, ObjectType
from signals_notebook.entities import Entity
from signals_notebook.entities.entity_store import EntityStore
from signals_notebook.exceptions import EIDError
from signals_notebook.materials import MaterialStore
from signals_notebook.materials.material import Material

CellContentType = TypeVar('CellContentType')


class ColumnDataType(str, Enum):
    NUMBER = 'number'
    INTEGER = 'integer'
    DATE_TIME = 'datetime'
    TEXT = 'text'
    LIST = 'list'
    MULTI_SELECT = 'multiSelect'
    ATTRIBUTE_LIST = 'attributeList'
    AUTOTEXT_LIST = 'autotextList'
    BOOLEAN = 'boolean'
    UNIT = 'unit'
    LINK = 'link'
    EXTERNAL_LINK = 'externalLink'


class ColumnDefinition(BaseModel):
    key: UUID
    title: str
    type: ColumnDataType
    is_external_key: Optional[bool] = Field(alias='isExternalKey', default=None)
    is_user_defined: Optional[bool] = Field(alias='isUserDefined', default=None)
    saved: Optional[bool] = Field(default=None)
    read_only: bool = Field(default=True, alias='readOnly')

    class Config:
        frozen = True


class AttributeListColumnDefinition(ColumnDefinition):
    type: Literal[ColumnDataType.ATTRIBUTE_LIST]
    options: List[str]
    attribute_list_eid: EID = Field(alias='attributeListEid')
    multi_select: bool = Field(alias='multiSelect')


class AutotextListColumnDefinition(ColumnDefinition):
    type: Literal[ColumnDataType.AUTOTEXT_LIST]
    options: List[str]
    autotext_list_eid: EID = Field(alias='autotextListEid')


class ListColumnDefinition(ColumnDefinition):
    type: Literal[ColumnDataType.LIST]
    options: List[str]


class MultiSelectColumnDefinition(ColumnDefinition):
    type: Literal[ColumnDataType.MULTI_SELECT]
    options: List[str]


class UnitColumnDefinition(ColumnDefinition):
    type: Literal[ColumnDataType.UNIT]
    measure: str
    default_unit: str = Field(alias='defaultUnit')


GenericColumnDefinition = Union[
    AttributeListColumnDefinition,
    AutotextListColumnDefinition,
    ListColumnDefinition,
    MultiSelectColumnDefinition,
    UnitColumnDefinition,
    ColumnDefinition,  # must be the last
]


class ColumnDefinitions(BaseModel):
    id: EID
    type: Literal[ObjectType.COLUMN_DEFINITIONS]
    columns: List[GenericColumnDefinition]

    class Config:
        frozen = True


class CellContent(GenericModel, Generic[CellContentType]):
    value: CellContentType
    values: Optional[List[CellContentType]] = None
    type: Optional[Union[EntityType, ObjectType, MaterialType]] = None
    display: Optional[str] = None

    class Config:
        validate_assignment = True


class CellContentDict(TypedDict):
    value: Any
    values: Optional[List[Any]]
    type: Optional[EntityType]
    display: Optional[str]


class UpdateCellRequest(GenericModel, Generic[CellContentType]):
    key: UUID
    content: CellContent[CellContentType]


class Cell(GenericModel, Generic[CellContentType]):
    id: UUID = Field(allow_mutation=False, alias='key')
    type: ColumnDataType = Field(allow_mutation=False)
    name: str = Field(allow_mutation=False)
    content: CellContent[CellContentType]
    _changed: bool = PrivateAttr(default=False)

    class Config:
        validate_assignment = True

    @property
    def value(self) -> Union[CellContentType, List[CellContentType]]:
        """Get content value

        Returns:
            Union[CellContentType, List[CellContentType]]
        """
        return self.content.values or self.content.value

    def _set_value(self, new_value: CellContentType, display: Optional[str] = None) -> None:
        self.content.value = new_value
        self.content.display = display

        self._changed = True

    @property
    def is_changed(self) -> bool:
        """Get is_changed value

        Returns:
            bool: True/False
        """
        return self._changed

    @property
    def display(self) -> str:
        """Get display field of content

        Returns:
            str
        """
        return self.content.display or ''

    @property
    def update_request(self) -> UpdateCellRequest[CellContentType]:
        """Get UpdateCellRequest

        Returns:
            UpdateCellRequest
        """
        return UpdateCellRequest[CellContentType](key=self.id, content=self.content)


class TextCell(Cell[str]):
    type: Literal[ColumnDataType.TEXT] = Field(allow_mutation=False)

    def set_value(self, new_value: str, display: Optional[str] = None) -> None:
        """Set new value to TextCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class NumberCell(Cell[float]):
    type: Literal[ColumnDataType.NUMBER] = Field(allow_mutation=False)

    def set_value(self, new_value: float, display: Optional[str] = None) -> None:
        """Set new value to NumberCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class IntegerCell(Cell[int]):
    type: Literal[ColumnDataType.INTEGER] = Field(allow_mutation=False)

    def set_value(self, new_value: int, display: Optional[str] = None) -> None:
        """Set new value to IntegerCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class BooleanCell(Cell[bool]):
    type: Literal[ColumnDataType.BOOLEAN] = Field(allow_mutation=False)

    def set_value(self, new_value: bool, display: Optional[str] = None) -> None:
        """Set new value to BooleanCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class DateTimeCell(Cell[DateTime]):
    type: Literal[ColumnDataType.DATE_TIME] = Field(allow_mutation=False)

    def set_value(self, new_value: DateTime, display: Optional[str] = None) -> None:
        """Set new value to DateTimeCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class ExternalLink(Cell[str]):
    type: Literal[ColumnDataType.EXTERNAL_LINK] = Field(allow_mutation=False)

    def set_value(self, new_value: str, display: Optional[str] = None) -> None:
        """Set new value to ExternalLink

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class LinkCell(Cell[Union[EID, MID]]):
    type: Literal[ColumnDataType.LINK] = Field(allow_mutation=False)

    @property
    def object(self) -> Union[Entity, Material]:
        """Get Material or Entity

        Returns:

        """
        object_id = self.content.value
        try:
            return MaterialStore.get(MID(object_id))
        except EIDError:
            pass

        return EntityStore.get(EID(object_id))

    def set_value(self, new_value: Union[EID, MID], display: str) -> None:
        """Set new value to LinkCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class UnitCell(Cell[float]):
    type: Literal[ColumnDataType.UNIT] = Field(allow_mutation=False)

    def set_value(self, new_value: float, display: Optional[str] = None) -> None:
        """Set new value to UnitCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class MultiSelectCell(Cell[str]):
    type: Literal[ColumnDataType.MULTI_SELECT] = Field(allow_mutation=False)
    content: CellContent[str]

    def set_value(self, new_value: Union[str, List[str]], display: Optional[str] = None) -> None:
        """Set new value to MultiSelectCell

        Args:
            new_value: new value or list of values of content value field
            display: new value of content display field

        Returns:

        """
        if isinstance(new_value, List):
            value = ', '.join(new_value)
            self.content.values = new_value
        else:
            value = new_value
            self.content.values = [new_value]

        super()._set_value(value, display)


class AttributeListCell(Cell[str]):
    type: Literal[ColumnDataType.ATTRIBUTE_LIST] = Field(allow_mutation=False)
    content: CellContent[str]

    def set_value(self, new_value: Union[str, List[str]], display: Optional[str] = None) -> None:
        """Set new value to AttributeListCell

        Args:
            new_value: new value or list of values of content value field
            display: new value of content display field

        Returns:

        """

        if isinstance(new_value, List):
            value = ', '.join(new_value)
            self.content.values = cast(List[str], new_value)
        else:
            value = new_value
            self.content.values = [new_value]

        super()._set_value(value, display)


class ListCell(Cell[str]):
    type: Literal[ColumnDataType.LIST] = Field(allow_mutation=False)

    def set_value(self, new_value: str, display: Optional[str] = None) -> None:
        """Set new value to ListCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


class AutotextListCell(Cell[str]):
    type: Literal[ColumnDataType.AUTOTEXT_LIST] = Field(allow_mutation=False)

    def set_value(self, new_value: str, display: Optional[str] = None) -> None:
        """Set new value to ListCell

        Args:
            new_value: new value of content value field
            display: new value of content display field

        Returns:

        """
        super()._set_value(new_value, display)


GenericCell = Annotated[
    Union[
        AttributeListCell,
        AutotextListCell,
        BooleanCell,
        DateTimeCell,
        ExternalLink,
        IntegerCell,
        LinkCell,
        ListCell,
        MultiSelectCell,
        NumberCell,
        TextCell,
        UnitCell,
    ],
    Field(discriminator='type'),
]



import json
import logging
from functools import cached_property
from typing import Any, cast, ClassVar, List, Literal, Optional

from pydantic import BaseModel, Field

from signals_notebook.common_types import Ancestors, EntityCreationRequestPayload, EntityType, Template
from signals_notebook.entities import Notebook
from signals_notebook.entities.container import Container
from signals_notebook.jinja_env import env
from signals_notebook.utils import FSHandler

CUSTOM_SYSTEM_OBJECT = 'Custom System Object'
log = logging.getLogger(__name__)


class _Attributes(BaseModel):
    name: str
    description: Optional[str] = None


class _Relationships(BaseModel):
    template: Optional[Template] = None
    ancestors: Optional[Ancestors] = None


class _Meta(BaseModel):
    ado_type_name: str = Field(alias='adoTypeName')

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True


class _RequestBody(BaseModel):
    type: EntityType
    meta: _Meta
    attributes: _Attributes
    relationships: Optional[_Relationships] = None


class _RequestPayload(EntityCreationRequestPayload[_RequestBody]):
    pass


class AdoType(BaseModel):
    id: str
    base_type: str = Field(alias='baseType')
    ado_name: str = Field(alias='adoName')

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True


class AdminDefinedObject(Container):
    type: Literal[EntityType.ADO] = Field(allow_mutation=False)
    ado: AdoType
    _template_name: ClassVar = 'ado.html'

    class Config:
        keep_untouched = (cached_property,)

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.ADO

    @classmethod
    def create(
        cls,
        *,
        name: str,
        ado_type_name: str,
        description: Optional[str] = None,
        template: Optional['AdminDefinedObject'] = None,
        notebook: Optional[Notebook] = None,
        digest: Optional[str] = None,
        force: bool = True,
    ) -> 'AdminDefinedObject':
        """Create new AdminDefinedObject in Signals Notebook

        Args:
            name: name of AdminDefinedObject
            ado_type_name: new type name for ADO object
            description: description of AdminDefinedObject
            template: AdminDefinedObject template
            notebook: notebook where create AdminDefinedObject
            digest: Indicate digest
            force: Force to create without doing digest check

        Returns:
            AdminDefinedObject
        """
        relationships = None
        if template or notebook:
            relationships = _Relationships(
                ancestors=Ancestors(data=[notebook.short_description]) if notebook else None,
                template=Template(data=template.short_description) if template else None,
            )
        request = _RequestPayload(
            data=_RequestBody(
                type=cls._get_entity_type(),
                meta=_Meta(ado_type_name=ado_type_name),
                attributes=_Attributes(
                    name=name,
                    description=description,
                ),
                relationships=relationships,
            )
        )

        log.debug('Creating AdminDefinedObject for: %s', cls.__name__)
        return cast('AdminDefinedObject', super()._create(digest=digest, force=force, request=request))

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        data = {
            'title': self.name,
            'description': self.description,
            'edited_at': self.edited_at,
            'children': self.get_children(),
        }

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)

    def dump(
        self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None
    ) -> None:  # type: ignore[override]
        """Dump AdminDefinedObject entity

        Args:
            base_path: content path where create dump
            fs_handler: FSHandler
            alias: Backup alias

        Returns:

        """
        metadata = {
            **self.ado.dict(exclude={'id'}),
            **{k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')},
        }
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata),
            base_alias=alias + [self.name, '__Metadata'] if alias else None,
        )
        for child in self.get_children():
            child.dump(fs_handler.join_path(base_path, self.eid), fs_handler)

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler, notebook: Notebook) -> None:
        """Load AdminDefinedObject entity

        Args:
            path: content path
            fs_handler: FSHandler
            notebook: Container where load Experiment entity

        Returns:

        """
        cls._load(path, fs_handler, notebook)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        from signals_notebook.item_mapper import ItemMapper

        metadata = json.loads(fs_handler.read(fs_handler.join_path(path, 'metadata.json')))
        experiment = cls.create(
            notebook=parent,
            name=metadata['name'],
            ado_type_name=metadata.get('ado_name', CUSTOM_SYSTEM_OBJECT),
            description=metadata['description'],
            force=True,
        )
        child_entities_folders = fs_handler.list_subfolders(path)
        for child_entity in child_entities_folders:
            child_entity_type = child_entity.split(':')[0]
            ItemMapper.get_item_class(child_entity_type)._load(
                fs_handler.join_path(path, child_entity), fs_handler, experiment
            )



from typing import ClassVar, Literal

from pydantic import Field

from signals_notebook.common_types import EntityType
from signals_notebook.entities import TodoList


class TaskContainer(TodoList):
    type: Literal[EntityType.TASK_CONTAINER] = Field(allow_mutation=False)  # type: ignore
    _template_name: ClassVar = 'task_container.html'

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.TASK_CONTAINER



from typing import List

import requests
from pydantic import BaseModel, parse_obj_as, PydanticValueError


class ErrorBody(BaseModel):
    status: str
    code: str
    title: str = ''
    detail: str = ''


class ErrorResponse(BaseModel):
    errors: List[ErrorBody]


class SignalsNotebookError(Exception):
    def __init__(self, response: requests.Response):
        """Handle Signals Notebook API errors

        Args:
            response: Response object, which contains a server's response to an HTTP request.
        """
        self.parsed_response = parse_obj_as(ErrorResponse, response.json())

    def __str__(self) -> str:
        error = self.parsed_response.errors[0]
        return (
            f'<SignalsNotebookError status={error.status} code={error.code}> title={error.title} detail={error.detail}>'
        )


class EIDError(PydanticValueError):
    msg_template = 'incorrect EID value: "{value}"'


class BulkExportJobAlreadyRunningError(Exception):
    def __init__(self, message=None):
        super().__init__(message)

    def __str__(self) -> str:
        return 'Only one Bulk Export job can be processed at a time'



import logging
from enum import Enum
from typing import ClassVar, Literal

from pydantic import Field

from signals_notebook.common_types import EntityType, File
from signals_notebook.entities import Entity
from signals_notebook.entities.container import Container
from signals_notebook.entities.contentful_entity import ContentfulEntity

log = logging.getLogger(__name__)


class Excel(ContentfulEntity):
    class ContentType(str, Enum):
        XLSX = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
        XLTX = 'application/vnd.openxmlformats-officedocument.spreadsheetml.template'
        XLSM = 'application/vnd.ms-excel.sheet.macroEnabled.12'
        XLTM = 'application/vnd.ms-excel.template.macroEnabled.12'
        XLS = 'application/vnd.ms-excel'
        CSV = 'text/csv'

    type: Literal[EntityType.EXCEL] = Field(allow_mutation=False)
    _template_name: ClassVar = 'excel.html'

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.EXCEL

    @classmethod
    def create(
        cls,
        *,
        container: Container,
        name: str,
        content_type: str = ContentType.XLSX,
        content: bytes = b'',
        force: bool = True,
    ) -> Entity:
        """Create Excel Entity

        Args:
            container: Container where create new Excel
            name: file name
            content_type: content type of Excel entity
            content: Excel content
            force: Force to post attachment

        Returns:
            Excel
        """
        cls.ContentType(content_type)
        log.debug('Create entity: %s with name: %s in Container: %s', cls.__name__, name, container.eid)
        return container.add_child(
            name=name,
            content=content,
            content_type=content_type,
            force=force,
        )

    def get_content(self) -> File:
        """Get Excel content

        Returns:
            File
        """
        return super()._get_content()



import abc
import cgi
import logging
from typing import cast, ClassVar, Optional, Union

from pydantic import BaseModel, Field

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import ChemicalDrawingFormat, EID, File, Response, ResponseData
from signals_notebook.entities.stoichiometry.cell import ColumnDefinition, ColumnDefinitions
from signals_notebook.entities.stoichiometry.data_grid import (
    Conditions,
    DataGridKind,
    DataGrids,
    Products,
    Reactants,
    Rows,
    Solvents,
)
from signals_notebook.jinja_env import env

log = logging.getLogger(__name__)


class ColumnDefinitionsResponse(Response[ColumnDefinitions]):
    pass


class StoichiometryDataResponse(Response[DataGrids]):
    pass


class Stoichiometry(BaseModel, abc.ABC):
    """Stoichiometry object of Signals Notebook"""

    eid: EID = Field(allow_mutation=False)
    reactants: Reactants = Field(default=Reactants(__root__=[]))
    products: Products = Field(default=Products(__root__=[]))
    solvents: Solvents = Field(default=Solvents(__root__=[]))
    conditions: Conditions = Field(default=Conditions(__root__=[]))
    _template_name: ClassVar = 'stoichiometry.html'

    class Config:
        validate_assignment = True

    @classmethod
    def set_template_name(cls, template_name: str) -> None:
        """Set new template name

        Args:
            template_name: name of the template (str)

        Returns:

        """
        log.debug('Setting new template for: %s...', cls.eid)
        cls._template_name = template_name
        log.debug('New template (%s) for %s was set', template_name, cls.eid)

    @classmethod
    def _get_endpoint(cls) -> str:
        return 'stoichiometry'

    @classmethod
    def _get_stoichiometry(cls, data: ResponseData) -> 'Stoichiometry':
        body = cast(DataGrids, data.body)

        stoichiometry = Stoichiometry(
            eid=data.eid,
            reactants=body.reactants,
            products=body.products,
            solvents=body.solvents,
            conditions=body.conditions,
        )

        for grid_kind in DataGridKind:
            grid_kind = cast(DataGridKind, grid_kind)
            grid = getattr(stoichiometry, grid_kind.value)
            grid = cast(Rows, grid)
            column_definitions = stoichiometry.get_column_definitions(grid_kind)
            grid.set_column_definitions(column_definitions)

        return stoichiometry

    @classmethod
    def fetch_data(cls, entity_eid: EID) -> Union['Stoichiometry', list['Stoichiometry']]:
        """Fetch stoichiometry data of experiment or chemicalDrawing by entity_id.
        Accepted entity types: experiment, chemicalDrawing.

        Args:
            entity_eid: Unique entity identifier

        Returns:
            Stoichiometry object or list of Stoichiometry objects
        """
        api = SignalsNotebookApi.get_default_api()
        fields = ', '.join(DataGridKind)
        log.debug('Fetching data for: %s...', entity_eid)

        response = api.call(
            method='GET',
            path=(cls._get_endpoint(), entity_eid),
            params={'fields': fields, 'value': 'normalized'},
        )

        result = StoichiometryDataResponse(**response.json())
        data = cast(ResponseData, result.data)

        if isinstance(data, list):
            stoichiometry_list = []
            for item in data:
                stoichiometry = cls._get_stoichiometry(data=item)
                stoichiometry_list.append(stoichiometry)
            log.debug('Returned stoichiometry data as list for: %s.', entity_eid)
            return stoichiometry_list
        else:
            stoichiometry = cls._get_stoichiometry(data=data)
            log.debug('Returned a single stoichiometry for: %s.', entity_eid)

            return stoichiometry

    def fetch_structure(self, row_id: str, format: Optional[ChemicalDrawingFormat] = None) -> File:
        """Fetch structure of reactants/products. Accepted entity types: chemicalDrawing.

        Args:
            row_id: row_id of grid data.
            format: one of the ChemicalDrawingFormat formats

        Returns:
            File
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Fetching structure for: %s, using row: (%s)...', self.eid, row_id)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.eid, row_id, 'structure'),
            params={'format': format},
        )

        content_disposition = response.headers.get('content-disposition', '')
        _, params = cgi.parse_header(content_disposition)

        log.debug('Return fetched structure as File for %s', self.eid)
        return File(
            name=params.get('filename', ''),
            content=response.content,
            content_type=response.headers.get('content-type', ''),
        )

    def get_column_definitions(self, data_grid_kind: DataGridKind) -> list[ColumnDefinition]:
        """Get column definitions of stoichiometry grid.

        Args:
            data_grid_kind (DataGridKind): kind of data grids in stoichiometry

        Returns:
            list of ColumnDefinition objects
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Getting column definitions for: %s...', self.eid)

        response = api.call(method='GET', path=(self._get_endpoint(), self.eid, 'columns', data_grid_kind))

        result = ColumnDefinitionsResponse(**response.json())
        body = cast(ResponseData, result.data).body

        log.debug('Column definitions for %s was returned', self.eid)
        return getattr(body, data_grid_kind, [])

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        data = {
            'reactants_html': self.reactants.get_html(),
            'products_html': self.products.get_html(),
            'solvents_html': self.solvents.get_html(),
            'conditions_html': self.conditions.get_html(),
        }

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)



from typing import Iterable, List, Optional, Protocol, Union


class FSHandler(Protocol):
    def write(self, path: str, data: Union[bytes, str], base_alias: Optional[Iterable[str]] = None):
        """Write file content into given path."""

    def read(self, path: str) -> bytes:
        """Return file content from given path."""

    def list_subfolders(self, path) -> List[str]:
        """Return subfolders names from given path."""

    @classmethod
    def join_path(cls, *paths: str) -> str:
        """Concatenate file paths."""



import logging
import mimetypes
from os.path import dirname, join

mime_types_file = join(dirname(__file__), 'mime.types')

mimetypes.init([mime_types_file])

logging.getLogger(__name__).addHandler(logging.NullHandler())



import logging
import mimetypes
import os
import re
from base64 import b64encode
from datetime import datetime
from enum import Enum
from typing import Any, Generic, List, Optional, TypeVar, Union
from uuid import UUID

from dateutil.parser import parse
from pydantic import BaseModel, Field, HttpUrl, validator
from pydantic.generics import GenericModel

from signals_notebook.exceptions import EIDError

EntityClass = TypeVar('EntityClass')
AnyModel = TypeVar('AnyModel')

log = logging.getLogger(__name__)


class ChemicalDrawingFormat(str, Enum):
    CDXML = 'cdxml'
    SVG = 'svg'
    MOL = 'mol'
    MOL3000 = 'mol-v3000'
    SMILES = 'smiles'


class ObjectType(str, Enum):
    ENTITY = 'entity'
    ADT_ROW = 'adtRow'
    COLUMN_DEFINITIONS = 'columnDefinitions'
    MATERIAL = 'material'
    ASSET_TYPE = 'assetType'
    ATTRIBUTE = 'attribute'
    STOICHIOMETRY = 'stoichiometry'
    PROPERTY = 'property'
    USER = 'user'
    PROFILE = 'profile'
    GROUP = 'group'
    ROLE = 'role'
    PLATE_ROW = 'plateRow'
    ATTRIBUTE_OPTION = 'option'
    CHOICE = 'choice'
    SUB_EXPERIMENT = 'subexpSummaryRow'
    CONTAINER = 'container'
    REACTION_PRODUCT = 'reactionProduct'
    REACTION_REACTANT = 'reactionReactant'
    REACTION_REAGENT = 'reactionReagent'


class EntityType(str, Enum):
    NOTEBOOK = 'journal'
    EXPERIMENT = 'experiment'
    TEXT = 'text'
    CHEMICAL_DRAWING = 'chemicalDrawing'
    GRID = 'grid'
    ASSET = 'asset'
    BIO_SEQUENCE = 'bioSequence'
    UPLOADED_RESOURCE = 'uploadedResource'
    IMAGE_RESOURCE = 'imageResource'
    WORD = 'viewonly'
    EXCEL = 'excel'
    SAMPLE = 'sample'
    SAMPLES_CONTAINER = 'samplesContainer'
    POWER_POINT = 'presentation'
    SPOTFIRE = 'spotfiredxp'
    TODO_LIST = 'linkedTaskContainer'
    TASK = 'task'
    PLATE_CONTAINER = 'plateContainer'
    MATERIAL_TABLE = 'materialsTable'
    PARALLEL_EXPERIMENT = 'paraexp'
    SUB_EXPERIMENT = 'parasubexp'
    SUB_EXPERIMENT_SUMMARY = 'paragrid'
    SUB_EXPERIMENT_LAYOUT = 'paraLayout'
    ADO = 'ado'
    REQUEST = 'request'
    TASK_CONTAINER = 'taskContainer'


class MaterialType(str, Enum):
    LIBRARY = 'assetType'
    ASSET = 'asset'
    BATCH = 'batch'


class EID(str):
    """Entity ID"""

    def __new__(cls, content: Any, validate: bool = True):
        if validate:
            cls.validate(content)
        return str.__new__(cls, content)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Any):
        """Validate Entity ID

        Args:
            v: Entity ID

        Returns:
        """
        if not isinstance(v, str):
            log.error('%s is not instance of str', v)
            raise EIDError(value=v)

        try:
            _type, _id, *_ = v.split(':')
            UUID(_id)
        except ValueError:
            log.exception('Cannot get id and type from value')
            raise EIDError(value=v)

        return cls(v, validate=False)

    @property
    def type(self) -> Union[EntityType, str]:
        """Get entity type

        Returns:
            One of the entity types
        """
        _type, _id, *_ = self.split(':')
        try:
            return EntityType(_type)
        except ValueError:
            log.exception('Cannot get type: %s. There is no the same type in program', _type)
            return _type

    @property
    def id(self) -> UUID:
        """Get UUID

        Returns:
            UUID
        """
        _type, _id, *_ = self.split(':')
        return UUID(_id)


class MID(str):
    """Material ID"""

    _id_pattern = re.compile('[0-9a-f]+', flags=re.IGNORECASE)

    def __new__(cls, content: Any, validate: bool = True):
        if validate:
            cls.validate(content)
        return str.__new__(cls, content)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Any):
        """Validate Material ID

        Args:
            v: Material ID

        Returns:

        """
        if not isinstance(v, str):
            log.error('%s is not instance of str', v)
            raise EIDError(value=v)

        try:
            _type, _id = v.split(':')
            MaterialType(_type)
        except ValueError:
            log.exception('Cannot get id and type from value')
            raise EIDError(value=v)

        if not cls._id_pattern.fullmatch(_id):
            log.error('ID: %s is not the same with %s pattern', _id, cls._id_pattern)
            raise EIDError(value=v)

        return cls(v, validate=False)

    @property
    def type(self) -> MaterialType:
        """Get one of the material types

        Returns:
            MaterialType
        """
        _type, _ = self.split(':')
        return MaterialType(_type)

    @property
    def id(self) -> str:
        """Get id of material type

        Returns:
            str id
        """
        _, _id = self.split(':')
        return _id


class AttrID(str):
    """Attribute ID"""

    def __new__(cls, content: Any, validate: bool = True):
        if validate:
            cls.validate(content)
        return str.__new__(cls, content)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Any):
        """Validate Attribute ID

        Args:
            v: Attribute ID

        Returns:

        """
        if not isinstance(v, str):
            log.error('%s is not instance of str', v)
            raise EIDError(value=v)

        try:
            _type, _id = v.split(':')
            int(_id)
        except ValueError:
            log.exception('Cannot get id and type from value')
            raise EIDError(value=v)

        if _type != ObjectType.ATTRIBUTE:
            log.error('Type: %s is not the same with %s', _type, ObjectType.ATTRIBUTE)
            raise EIDError(value=v)

        return cls(v, validate=False)

    @property
    def type(self) -> ObjectType:
        """Get one of the object types

        Returns:
            ObjectType
        """
        _type, _ = self.split(':')
        return ObjectType(_type)

    @property
    def id(self) -> int:
        """Get id of object type

        Returns:
            int id
        """
        _, _id = self.split(':')
        return int(_id)


class Links(BaseModel):
    self: HttpUrl
    first: Optional[HttpUrl] = None
    next: Optional[HttpUrl] = None
    prev: Optional[HttpUrl] = None

    @validator('*', pre=True)
    def escape_spaces(cls, v: Optional[str]) -> Optional[str]:
        """Replace all spaces

        Args:
            v: value

        Returns:
            value with replaced spaces
        """
        if v is not None:
            return v.replace(' ', '%20')

        return v


class ResponseData(GenericModel, Generic[EntityClass]):
    type: ObjectType
    eid: Union[EID, MID, AttrID, UUID, str] = Field(alias='id')
    links: Optional[Links] = None
    body: EntityClass = Field(alias='attributes')
    relationships: Optional[dict[str, Any]] = Field(default=None)
    meta: Optional[dict[str, Any]] = Field(default=None)

    def __init__(self, _context: Optional[dict[str, Any]] = None, **kwargs):
        attributes = kwargs.get('attributes', {})

        if _context:
            attributes = {**attributes, **_context}

        super().__init__(**{**kwargs, 'attributes': attributes})


class Response(GenericModel, Generic[EntityClass]):
    links: Optional[Links] = None
    data: Union[ResponseData[EntityClass], List[ResponseData[EntityClass]]]

    def __init__(self, _context: Optional[dict[str, Any]] = None, **kwargs):
        data = kwargs.get('data', {})

        if _context:
            if isinstance(data, list):
                data = [{'_context': _context, **item} for item in data]
            else:
                data = {'_context': _context, **data}

        super().__init__(**{**kwargs, 'data': data})


class DataObject(GenericModel, Generic[AnyModel]):
    data: AnyModel


class DataList(GenericModel, Generic[AnyModel]):
    data: List[AnyModel]


class EntityCreationRequestPayload(DataObject[AnyModel], Generic[AnyModel]):
    pass


class EntityShortDescription(BaseModel):
    type: Union[EntityType, str]
    id: EID


class Template(DataObject[EntityShortDescription]):
    pass


class Ancestors(DataList[EntityShortDescription]):
    pass


class File(BaseModel):
    name: str
    content: bytes
    content_type: str

    def __init__(self, f=None, **kwargs):
        if f:
            name = os.path.basename(f.name)
            content = f.read()
            content_type, _ = mimetypes.guess_type(name)

            super().__init__(name=name, content=content, content_type=content_type)
        else:
            super().__init__(**kwargs)

    @property
    def size(self) -> int:
        """Get file size

        Returns:
            file size
        """
        return len(self.content)

    @property
    def base64(self) -> bytes:
        return b64encode(self.content)

    @classmethod
    def read(cls, file_name: str, mode='rb') -> 'File':
        """Read content of the file

        Args:
            file_name: file name in string format
            mode: specifies the mode in which the file is opened

        Returns:
            File
        """
        with open(file_name, mode) as f:
            return cls(f)

    def save(self, path: str) -> None:
        """Save content in file

        Args:
            path: path to the file

        Returns:

        """
        _path = path
        if os.path.isdir(path):
            _path = os.path.join(path, self.name)

        with open(_path, 'wb') as f:
            f.write(self.content)


class DateTime(datetime):
    @classmethod
    def __get_validators__(cls):
        yield cls._validate_date

    @staticmethod
    def _validate_date(value: Union[str, datetime]) -> datetime:
        if isinstance(value, datetime):
            return value

        return parse(value)



import json
import logging
from datetime import datetime
from typing import cast, Generator, Literal

from pydantic import BaseModel, Field

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import ObjectType, Response, ResponseData
from signals_notebook.users.user import User


log = logging.getLogger(__name__)


class Group(BaseModel):
    type: Literal[ObjectType.GROUP] = Field(allow_mutation=False)
    id: str
    is_system: bool = Field(alias='isSystem', allow_mutation=False)
    name: str
    description: str
    created_at: datetime = Field(alias='createdAt', allow_mutation=False)
    edited_at: datetime = Field(alias='editedAt', allow_mutation=False)
    digest: str

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True

    @classmethod
    def _get_endpoint(cls) -> str:
        return 'groups'

    @classmethod
    def get_list(cls) -> Generator['Group', None, None]:
        """Get all groups

        Returns:
            Group
        """
        log.debug('Get List of Groups')

        api = SignalsNotebookApi.get_default_api()
        response = api.call(
            method='GET',
            path=(cls._get_endpoint(),),
        )
        result = GroupResponse(**response.json())
        yield from [cast(ResponseData, item).body for item in result.data]

        while result.links and result.links.next:
            response = api.call(
                method='GET',
                path=result.links.next,
            )

            result = GroupResponse(**response.json())
            yield from [cast(ResponseData, item).body for item in result.data]

        log.debug('List of Groups were got successfully.')

    @classmethod
    def get(cls, group_id: str) -> 'Group':
        """Get group by id

        Args:
            group_id: Unique user group identifier
        Returns:
            Group
        """
        log.debug('Get Group: %s', group_id)

        api = SignalsNotebookApi.get_default_api()
        response = api.call(
            method='GET',
            path=(cls._get_endpoint(), group_id),
        )
        result = GroupResponse(**response.json())

        log.debug('Group: %s was got successfully.', group_id)

        return cast(ResponseData, result.data).body

    @classmethod
    def create(cls, is_system: bool, name: str, description: str) -> 'Group':
        """Create new Group

        Returns:
            Group
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Create Group: %s...', cls.__name__)

        response = api.call(
            method='POST',
            path=(cls._get_endpoint(),),
            json={'data': {'attributes': {'name': name, 'description': description, 'isSystem': is_system}}},
        )
        log.debug('Group: %s was created.', cls.__name__)

        result = GroupResponse(**response.json())
        return cast(ResponseData, result.data).body

    def save(self, force: bool = True) -> None:
        """Update attributes of a specified group.

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()
        api.call(
            method='PATCH',
            path=(
                self._get_endpoint(),
                self.id,
            ),
            params={
                'force': json.dumps(force),
            },
            json={
                'data': {
                    'attributes': self.dict(
                        include={
                            'name',
                            'description',
                        },
                    )
                },
            },
        )
        log.debug('Group: %s was saved successfully', self.id)

    def delete(self):
        """Delete user group by id.

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Disable Group: %s...', self.id)

        api.call(
            method='DELETE',
            path=(self._get_endpoint(), self.id),
        )
        log.debug('Group: %s was disabled successfully', self.id)

    def get_members(self) -> list[User]:
        """Get user group members

        Returns:
            Users
        """
        log.debug('Get Group Members')

        api = SignalsNotebookApi.get_default_api()
        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.id, 'members'),
        )

        log.debug('Group members were got successfully.')
        return [User.get(user['id']) for user in response.json()['data']]

    def add_user(self, user: User, force: bool = True) -> list[User]:
        """Add user to user group

        Args:
            user: User object
            force: Force to update

        Returns:
            list[User]
        """
        log.debug('Adding members to group')

        api = SignalsNotebookApi.get_default_api()

        api.call(
            method='POST',
            path=(self._get_endpoint(), self.id, 'members'),
            params={
                'force': json.dumps(force),
            },
            json={
                'data': {
                    'attributes': {'userId': user.id},
                }
            },
        )

        log.debug('Group member: %s was added successfully', user.id)

        return self.get_members()

    def delete_user(self, user: User) -> None:
        """Delete user from user group.

        Args:
            user: User object

        Returns:

        """
        log.debug('Deleting members from group')

        api = SignalsNotebookApi.get_default_api()

        api.call(
            method='DELETE',
            path=(self._get_endpoint(), self.id, 'members', user.id),
        )
        log.debug('Group member: %s was deleted successfully', user.id)


class GroupResponse(Response[Group]):
    pass



import logging
from typing import cast, Generator, Optional

from pydantic import BaseModel, Field

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import Response, ResponseData


log = logging.getLogger(__name__)


class Privelege(BaseModel):
    can_move_experiments: bool = Field(alias='canMoveExperiments', default=True)
    can_manage_system_templates: bool = Field(alias='canManageSystemTemplates', default=True)
    can_trash_requests: bool = Field(alias='canTrashRequests', default=True)
    can_trash_samples: bool = Field(alias='canTrashSamples', default=True)
    can_share: bool = Field(alias='canShare', default=True)
    can_add_materials: bool = Field(alias='canAddMaterials', default=True)
    can_trash_experiments: bool = Field(alias='canTrashExperiments', default=True)
    can_trash_materials: bool = Field(alias='canTrashMaterials', default=True)
    can_trash_notebooks: bool = Field(alias='canTrashNotebooks', default=True)
    can_view_materials: bool = Field(alias='canViewMaterials', default=True)
    can_manage_material_libraries: bool = Field(alias='canManageMaterialLibraries', default=True)
    can_share_templates: bool = Field(alias='canShareTemplates', default=True)
    can_manage_attributes: bool = Field(alias='canManageAttributes', default=True)
    can_manage_groups: bool = Field(alias='canManageGroups', default=True)
    can_search_eln_archive: bool = Field(alias='canSearchElnArchive', default=True)
    can_configure: bool = Field(alias='canConfigure', default=True)
    can_edit_materials: bool = Field(alias='canEditMaterials', default=True)

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True


class Role(BaseModel):
    id: str
    name: str
    description: Optional[str]
    privileges: Optional[Privelege]

    @classmethod
    def _get_endpoint(cls) -> str:
        return 'roles'

    @classmethod
    def get_list(cls) -> Generator['Role', None, None]:
        log.debug('Get List of Roles')

        api = SignalsNotebookApi.get_default_api()
        response = api.call(
            method='GET',
            path=(cls._get_endpoint(),),
        )
        result = RoleResponse(**response.json())
        yield from [cast(ResponseData, item).body for item in result.data]

        log.debug('List of Roles were got successfully.')

    @classmethod
    def get(cls, role_id: str) -> 'Role':
        """Get role by id

        Args:
            role_id: Unique role identifier
        Returns:
            Role
        """
        log.debug('Getting Role: ', role_id)

        api = SignalsNotebookApi.get_default_api()
        response = api.call(
            method='GET',
            path=(cls._get_endpoint(), role_id),
        )
        result = RoleResponse(**response.json())

        log.debug('Role: %s was got successfully.', role_id)

        return cast(ResponseData, result.data).body


class RoleResponse(Response[Role]):
    pass



import logging
from enum import Enum
from operator import attrgetter
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, Field, PrivateAttr

from signals_notebook.common_types import ObjectType
from signals_notebook.entities.tables.cell import Cell, GenericCell, UpdateCellRequest

log = logging.getLogger(__name__)


class RowAction(str, Enum):
    UPDATE = 'update'
    CREATE = 'create'
    DELETE = 'delete'


class DeleteRowActionBody(BaseModel):
    action: Literal[RowAction.DELETE] = Field(allow_mutation=False, default=RowAction.DELETE)

    class Config:
        validate_assignment = True


class UpdateRowActionBody(BaseModel):
    action: Literal[RowAction.UPDATE] = Field(allow_mutation=False, default=RowAction.UPDATE)
    cells: List[UpdateCellRequest]

    class Config:
        validate_assignment = True


class CreateRowActionBody(BaseModel):
    action: Literal[RowAction.CREATE] = Field(allow_mutation=False, default=RowAction.CREATE)
    cells: List[UpdateCellRequest]

    class Config:
        validate_assignment = True


class ChangeRowRequest(BaseModel):
    type: Literal[ObjectType.ADT_ROW] = Field(allow_mutation=False, default=ObjectType.ADT_ROW)

    class Config:
        validate_assignment = True


class DeleteRowRequest(ChangeRowRequest):
    id: UUID
    body: DeleteRowActionBody = Field(alias='attributes', default_factory=DeleteRowActionBody)


class UpdateRowRequest(ChangeRowRequest):
    id: UUID
    body: UpdateRowActionBody = Field(alias='attributes')


class CreateRowRequest(ChangeRowRequest):
    body: CreateRowActionBody = Field(alias='attributes')


class Row(BaseModel):
    id: Optional[UUID] = Field(allow_mutation=False, default=None)
    type: Literal[ObjectType.ADT_ROW] = Field(allow_mutation=False, default=ObjectType.ADT_ROW)
    cells: List[GenericCell]
    _cells_dict: Dict[Union[UUID, str], GenericCell] = PrivateAttr(default={})
    _deleted: bool = PrivateAttr(default=False)

    class Config:
        validate_assignment = True

    def __init__(self, **data):
        super().__init__(**data)

        for cell in self.cells:
            self._cells_dict[cell.id] = cell
            self._cells_dict[cell.name] = cell

    def get(self, value: Union[str, UUID], default: Any = None) -> Union[Cell, Any]:
        """Get one of the GenericCell objects by value

        Args:
            value: key to get one of the GenericCell objects
            default: default value if key doesn't exist

        Returns:
            Union[Cell, Any]
        """
        try:
            return self[value]
        except KeyError:
            log.debug('KeyError were caught. Default value returned')
            return default

    @property
    def is_deleted(self) -> bool:
        """Get is_deleted field

        Returns:
            bool: True/False
        """
        return self._deleted

    @property
    def is_changed(self) -> bool:
        """Get is_changed field

        Returns:
            bool: True/False
        """
        return any([cell.is_changed for cell in self.cells])

    @property
    def is_new(self) -> bool:
        """Check if id field exists

        Returns:
            bool: True/False
        """
        return self.id is None

    def get_values(self, use_labels: bool = True) -> Dict[str, Any]:
        """Get row values

        Args:
            use_labels: use cels names

        Returns:
            Dict[str, Any]
        """
        key_getter = attrgetter('name') if use_labels else attrgetter('key')
        return {key_getter(cell): cell.value for cell in self.cells}

    def __getitem__(self, index: Union[int, str, UUID]) -> GenericCell:
        if isinstance(index, int):
            return self.cells[index]

        if isinstance(index, str):
            if index in self._cells_dict:
                return self._cells_dict[index]

            try:
                if UUID(index) in self._cells_dict:
                    return self._cells_dict[UUID(index)]
            except ValueError:
                pass

        if isinstance(index, UUID):
            return self._cells_dict[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        return self.cells.__iter__()

    def delete(self) -> None:
        """Delete Row

        Returns:

        """
        self._deleted = True

    def get_change_request(self) -> Optional[ChangeRowRequest]:
        """Get ChangeRowRequest depending on Row status

        Returns:
            Optional[ChangeRowRequest]
        """
        if self.is_deleted:
            return DeleteRowRequest(id=self.id)

        if self.is_changed:
            return UpdateRowRequest(
                id=self.id,
                attributes=UpdateRowActionBody(cells=[cell.update_request for cell in self.cells if cell.is_changed]),
            )

        if self.is_new:
            return CreateRowRequest(attributes=CreateRowActionBody(cells=[cell.update_request for cell in self.cells]))

        return None



from enum import Enum
from typing import Generic, List, Optional, TypeVar
from uuid import UUID

from pydantic import Field
from pydantic.generics import GenericModel


CellContentType = TypeVar('CellContentType')


class ColumnDataType(str, Enum):
    NUMBER = 'NUMBER'
    INTEGER = 'INTEGER'
    DATE_TIME = 'DATETIME'
    TEXT = 'TEXT'
    LIST = 'LIST'
    MULTI_SELECT = 'MULTI_SELECT'
    ATTRIBUTE_LIST = 'ATTRIBUTE_LIST'
    AUTOTEXT_LIST = 'AUTOTEXT_LIST'
    BOOLEAN = 'BOOLEAN'
    UNIT = 'UNIT'
    LINK = 'LINK'
    EXTERNAL_LINK = 'EXTERNAL_LINK'


class CellContent(GenericModel, Generic[CellContentType]):
    user: Optional[str]
    value: Optional[CellContentType] = None
    values: Optional[List[CellContentType]] = None

    class Config:
        validate_assignment = True


class PlateCell(GenericModel, Generic[CellContentType]):
    id: UUID = Field(allow_mutation=False, alias='key')
    type: ColumnDataType = Field(allow_mutation=False)
    name: str = Field(allow_mutation=False)
    content: CellContent[CellContentType]

    class Config:
        validate_assignment = True



from datetime import datetime
from typing import Optional

from pydantic import Field

from signals_notebook.base import PatchedModel
from signals_notebook.common_types import MID


class BaseMaterialEntity(PatchedModel):
    asset_type_id: str = Field(alias='assetTypeId', allow_mutation=False)
    eid: MID = Field(allow_mutation=False)
    library_name: str = Field(allow_mutation=False, alias='library')
    digest: Optional[str] = Field(allow_mutation=False, default=None)
    name: str = Field(title='Name', allow_mutation=False)
    created_at: datetime = Field(alias='createdAt', allow_mutation=False)
    edited_at: datetime = Field(alias='editedAt', allow_mutation=False)

    class Config:
        validate_assignment = True
        allow_population_by_field_name = True

    def __str__(self) -> str:
        return f'<{self.__class__.__name__} eid={self.eid}>'

    @classmethod
    def _get_endpoint(cls) -> str:
        return 'materials'



import abc
import json
import logging
import mimetypes
import os
from typing import cast, Generator, List, Optional, Union

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import EntityType, Response, ResponseData
from signals_notebook.entities import Entity
from signals_notebook.utils.fs_handler import FSHandler

log = logging.getLogger(__name__)


class Container(Entity, abc.ABC):
    @classmethod
    @abc.abstractmethod
    def _get_entity_type(cls) -> EntityType:
        pass

    def add_child(
        self,
        name: str,
        content: bytes,
        content_type: Optional[str] = None,
        force: bool = True,
    ) -> Entity:
        """Upload a file to an entity as a child.

        Args:
            name: file name
            content: entity content
            content_type: entity type
            force: Force to post attachment

        Returns:

        """
        api = SignalsNotebookApi.get_default_api()

        if content_type:
            extension = mimetypes.guess_extension(content_type)
            file_name = f'{os.path.splitext(name)[0]}{extension}' if extension else name
        else:
            content_type = mimetypes.guess_type(name)[0]
            file_name = name

        response = api.call(
            method='POST',
            path=(self._get_endpoint(), self.eid, 'children', file_name),
            params={
                'digest': None if force else self.digest,
                'force': json.dumps(force),
            },
            headers={
                'Content-Type': content_type or 'application/octet-stream',
            },
            data=content,
        )
        log.debug('Added child: %s to Container: %s', self.name, self.eid)

        entity_classes = (*Entity.get_subclasses(), Entity)
        result = Response[Union[entity_classes]](**response.json())  # type: ignore

        return cast(ResponseData, result.data).body

    def get_children(self, order: Optional[str] = None) -> Generator[Entity, None, None]:
        """Get children of a specified entity.

        Returns:
            list of Entities
        """
        api = SignalsNotebookApi.get_default_api()
        log.debug('Get children for: %s', self.eid)

        params = {'order': order} if order else {}
        response = api.call(method='GET', path=(self._get_endpoint(), self.eid, 'children'), params=params)

        entity_classes = (*Entity.get_subclasses(), Entity)

        result = Response[Union[entity_classes]](**response.json())  # type: ignore

        yield from [cast(ResponseData, item).body for item in result.data]

        while result.links and result.links.next:
            response = api.call(
                method='GET',
                path=result.links.next,
            )

            result = Response[Union[entity_classes]](**response.json())  # type: ignore
            yield from [cast(ResponseData, item).body for item in result.data]

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        metadata = {k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')}
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata),
            base_alias=alias + [self.name, '__Metadata'] if alias else None,
        )
        for child in self.get_children():
            child.dump(fs_handler.join_path(base_path, self.eid), fs_handler, alias + [self.name] if alias else None)

    @classmethod
    def dump_templates(cls, base_path: str, fs_handler: FSHandler) -> None:
        """Dump Container templates

        Args:
            base_path: content path where create templates dump
            fs_handler: FSHandler

        Returns:

        """
        from signals_notebook.entities import EntityStore

        entity_type = cls._get_entity_type()

        templates = EntityStore.get_list(
            include_types=[entity_type], include_options=[EntityStore.IncludeOptions.TEMPLATE]
        )

        for template in templates:
            template.dump(
                fs_handler.join_path(base_path, 'templates', entity_type),
                fs_handler,
                ['Templates', entity_type.value],
            )



import json
import logging
from enum import Enum
from functools import cached_property
from typing import Any, cast, ClassVar, Dict, Generator, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from signals_notebook.common_types import Ancestors, EntityCreationRequestPayload, EntityType, Template
from signals_notebook.entities import Entity
from signals_notebook.entities.container import Container
from signals_notebook.entities.notebook import Notebook
from signals_notebook.entities.stoichiometry.stoichiometry import Stoichiometry
from signals_notebook.jinja_env import env
from signals_notebook.utils.fs_handler import FSHandler

log = logging.getLogger(__name__)


class _Attributes(BaseModel):
    name: Optional[str] = None
    description: Optional[str] = None
    organization: Optional[str] = None
    department: Optional[str] = None
    project: Optional[str] = None
    modality: Optional[str] = None


class _Relationships(BaseModel):
    template: Optional[Template] = None
    ancestors: Optional[Ancestors] = None


class _RequestBody(BaseModel):
    type: EntityType
    attributes: _Attributes
    relationships: Optional[_Relationships] = None


class _RequestPayload(EntityCreationRequestPayload[_RequestBody]):
    pass


class ExperimentState(str, Enum):
    OPEN = 'open'
    CLOSED = 'closed'


class Experiment(Container):
    type: Literal[EntityType.EXPERIMENT] = Field(allow_mutation=False)
    state: Optional[ExperimentState] = Field(allow_mutation=False, default=None)
    _template_name: ClassVar = 'experiment.html'

    class Config:
        keep_untouched = (cached_property,)

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.EXPERIMENT

    @classmethod
    def create(
        cls,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        template: Optional['Experiment'] = None,
        notebook: Optional[Notebook] = None,
        digest: Optional[str] = None,
        force: bool = True,
        attributes: Optional[Dict[str, Any]] = None,
    ) -> 'Experiment':
        """Create new Experiment in Signals Notebook

        Args:
            name: name of experiment
            description: description of experiment
            template: experiment template
            notebook: notebook where create experiment
            digest: Indicate digest
            force: Force to create without doing digest check
            attributes:

        Returns:
            Experiment
        """

        relationships = None
        if template or notebook:
            relationships = _Relationships(
                ancestors=Ancestors(data=[notebook.short_description]) if notebook else None,
                template=Template(data=template.short_description) if template else None,
            )

        request = _RequestPayload(
            data=_RequestBody(
                type=cls._get_entity_type(),
                attributes=_Attributes(name=name, description=description, **(attributes or {})),
                relationships=relationships,
            )
        )

        log.debug('Creating Notebook for: %s', cls.__name__)
        return cast('Experiment', super()._create(digest=digest, force=force, request=request))

    @cached_property
    def stoichiometry(self) -> Union[Stoichiometry, list[Stoichiometry]]:
        """Fetch stoichiometry data of experiment

        Returns:
            Stoichiometry object or list of Stoichiometry objects
        """
        log.debug('Fetching data in Stoichiometry for: %s', self.eid)
        return Stoichiometry.fetch_data(self.eid)

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        data = {
            'title': self.name,
            'description': self.description,
            'edited_at': self.edited_at,
            'state': self.state,
            'children': self.get_children(),
        }

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)

    def get_children(self, order: Optional[str] = 'layout') -> Generator[Entity, None, None]:
        """Get children of Experiment.

        Returns:
            list of Entities
        """
        return super().get_children(order=order)

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler, notebook: Notebook) -> None:
        """Load Experiment entity

        Args:
            path: content path
            fs_handler: FSHandler
            notebook: Container where load Experiment entity

        Returns:

        """
        cls._load(path, fs_handler, notebook)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        from signals_notebook.item_mapper import ItemMapper

        metadata = json.loads(fs_handler.read(fs_handler.join_path(path, 'metadata.json')))
        try:
            experiment = cls.create(
                notebook=parent,
                name=metadata['name'],
                description=metadata['description'],
                force=True,
                attributes=dict(
                    organization=metadata['Organization'],
                    project=metadata['Project'],
                    modality=metadata['Modality'],
                    department=metadata['Department'],
                ),
            )
        except Exception as e:
            log.error(str(e))
            if 'According to template, name is auto generated, can not be specified' in str(e):
                log.error('Retrying create')
                experiment = cls.create(
                    notebook=parent,
                    description=metadata['description'],
                    force=True,
                    attributes=dict(
                        organization=metadata['Organization'],
                        project=metadata['Project'],
                        modality=metadata['Modality'],
                        department=metadata['Department'],
                    ),
                )
            else:
                raise e
        child_entities_folders = fs_handler.list_subfolders(path)
        for child_entity in child_entities_folders:
            child_entity_type = child_entity.split(':')[0]
            try:
                ItemMapper.get_item_class(child_entity_type)._load(
                    fs_handler.join_path(path, child_entity), fs_handler, experiment
                )
            except NotImplementedError:
                log.error('Failed to load entity %s. Not supported' % child_entity_type)

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        metadata = {k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')}
        self._reload_properties()
        for prop in self._properties:
            if prop.name in ('Department', 'Project', 'Modality', 'Organization'):
                metadata[prop.name] = prop.value

        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata),
            alias + [self.name, '__Metadata'] if alias else None,
        )
        for child in self.get_children():
            try:
                child.dump(
                    fs_handler.join_path(base_path, self.eid), fs_handler, alias + [self.name] if alias else None
                )
            except Exception as e:
                log.error(str(e))



import logging
from enum import Enum
from typing import ClassVar, Literal

from pydantic import Field

from signals_notebook.common_types import EntityType, File
from signals_notebook.entities import Entity
from signals_notebook.entities.container import Container
from signals_notebook.entities.contentful_entity import ContentfulEntity
from signals_notebook.jinja_env import env

log = logging.getLogger(__name__)


class BiologicalSequence(ContentfulEntity):
    class ContentType(str, Enum):
        FASTA = 'biosequence/fasta'
        GB = 'biosequence/genbank'
        GBK = 'biosequence/genpept'
        SW = 'chemical/x-swissprot'
        DNA = 'application/vnd.snapgene.dna'
        PROT = 'application/vnd.snapgene.protein'

    type: Literal[EntityType.BIO_SEQUENCE] = Field(allow_mutation=False)
    _template_name: ClassVar = 'bio_sequence.html'

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.BIO_SEQUENCE

    @classmethod
    def create(
        cls,
        *,
        container: Container,
        name: str,
        content: bytes = b'',
        content_type: str = ContentType.GB,
        force: bool = True,
    ) -> Entity:
        """Create BiologicalSequence Entity

        Args:
            container: Container where create new BiologicalSequence
            name: file name
            content_type: content type of BiologicalSequence entity
            content: BiologicalSequence content
            force: Force to post attachment

        Returns:
            BiologicalSequence
        """
        cls.ContentType(content_type)
        log.debug('Create entity: %s with name: %s in Container: %s', cls.__name__, name, container.eid)
        return container.add_child(
            name=name,
            content=content,
            content_type=content_type,
            force=force,
        )

    def get_content(self) -> File:
        """Get BiologicalSequence content

        Returns:
            File
        """
        return super()._get_content()

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """
        data = {'name': self.name}
        file = self.get_content()
        data['bio_sequence'] = 'data:{};base64,{}'.format(file.content_type, file.base64.decode('ascii'))

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(data=data)



import cgi
import json
import logging
from enum import Enum
from typing import Any, cast, Dict, List, Literal, Optional, Union
from uuid import UUID

import pandas as pd
from pydantic import Field, PrivateAttr

from signals_notebook.api import SignalsNotebookApi
from signals_notebook.common_types import DataList, EntityType, File, Response, ResponseData
from signals_notebook.entities import Entity, EntityStore
from signals_notebook.entities.container import Container
from signals_notebook.entities.tables.cell import Cell, CellContentDict, ColumnDefinitions, GenericColumnDefinition
from signals_notebook.entities.tables.row import ChangeRowRequest, Row
from signals_notebook.jinja_env import env
from signals_notebook.utils import FSHandler

log = logging.getLogger(__name__)


class TableDataResponse(Response[Row]):
    pass


class ColumnDefinitionsResponse(Response[ColumnDefinitions]):
    pass


class ChangeTableDataRequest(DataList[ChangeRowRequest]):
    pass


class Table(Entity):
    class ContentType(str, Enum):
        JSON = 'application/json'
        CSV = 'text/csv'

    type: Literal[EntityType.GRID] = Field(allow_mutation=False)
    _rows: List[Row] = PrivateAttr(default=[])
    _rows_by_id: Dict[UUID, Row] = PrivateAttr(default={})
    _template_name = 'table.html'

    @classmethod
    def _get_entity_type(cls) -> EntityType:
        return EntityType.GRID

    @classmethod
    def _get_adt_endpoint(cls) -> str:
        return 'adt'

    def _reload_data(self) -> None:
        api = SignalsNotebookApi.get_default_api()
        log.debug('Reloading data in Table: %s...', self.eid)

        response = api.call(
            method='GET',
            path=(self._get_adt_endpoint(), self.eid),
            params={
                'value': 'normalized',
            },
        )

        result = TableDataResponse(**response.json())

        self._rows = []
        self._rows_by_id = {}
        for item in result.data:
            row = cast(Row, cast(ResponseData, item).body)
            assert row.id

            self._rows.append(row)
            self._rows_by_id[row.id] = row
        log.debug('Data in Table: %s were reloaded', self.eid)

    def get_column_definitions_list(self) -> List[GenericColumnDefinition]:
        """Fetch column definitions

        Returns:
            List[GenericColumnDefinition]
        """
        api = SignalsNotebookApi.get_default_api()

        response = api.call(method='GET', path=(self._get_adt_endpoint(), self.eid, '_column'))

        result = ColumnDefinitionsResponse(**response.json())

        return cast(ResponseData, result.data).body.columns

    def get_column_definitions_map(self) -> Dict[str, GenericColumnDefinition]:
        """Get column definitions as a dictionary

        Returns:
            Dict[str, GenericColumnDefinition]
        """
        column_definitions = self.get_column_definitions_list()
        column_definitions_map: Dict[str, GenericColumnDefinition] = {}

        for column_definition in column_definitions:
            column_definitions_map[str(column_definition.key)] = column_definition
            column_definitions_map[column_definition.title] = column_definition

        return column_definitions_map

    def as_dataframe(self, use_labels: bool = True) -> pd.DataFrame:
        """Get as data table

        Args:
            use_labels: use cells names

        Returns:
            pd.DataFrame
        """
        if not self._rows:
            self._reload_data()

        data = []
        index = []
        for row in self._rows:
            index.append(row.id)
            data.append(row.get_values(use_labels))

        return pd.DataFrame(data=data, index=index)

    def as_raw_data(self, use_labels: bool = True) -> List[Dict[str, Any]]:
        """Get as a list of dictionaries

        Args:
            use_labels: use cels names

        Returns:

        """
        if not self._rows:
            self._reload_data()

        data = []
        for row in self._rows:
            data.append(row.get_values(use_labels))

        return data

    def __getitem__(self, index: Union[int, str, UUID]) -> Row:
        if not self._rows:
            self._reload_data()

        if isinstance(index, int):
            return self._rows[index]

        if isinstance(index, str):
            return self._rows_by_id[UUID(index)]

        if isinstance(index, UUID):
            return self._rows_by_id[index]

        raise IndexError('Invalid index')

    def __iter__(self):
        if not self._rows:
            self._reload_data()

        return self._rows.__iter__()

    def delete_row_by_id(self, row_id: Union[str, UUID], digest: Optional[str] = None, force: bool = True) -> None:
        """

        Args:
            row_id: id of the row
            digest: Indicate digest of entity. It is used to avoid conflict while concurrent editing.
            If the parameter 'force' is true, this parameter is optional.
            If the parameter 'force' is false, this parameter is required.
            force: Force to update properties without digest check.

        Returns:

        """
        if isinstance(row_id, UUID):
            _row_id = row_id.hex
        else:
            _row_id = row_id

        api = SignalsNotebookApi.get_default_api()

        api.call(
            method='DELETE',
            path=(self._get_adt_endpoint(), self.eid, _row_id),
            params={
                'digest': digest,
                'force': json.dumps(force),
            },
        )

        log.debug('Row %s was deleted', row_id)
        self._reload_data()

    def add_row(self, data: Dict[str, CellContentDict]) -> None:
        """Add row in the table

        Args:
            data: Cells to add in the row

        Returns:

        """
        column_definitions_map = self.get_column_definitions_map()

        prepared_data: List[Dict[str, Any]] = []
        for key, value in data.items():
            column_definition = column_definitions_map.get(key)
            if not column_definition:
                continue

            prepared_data.append(
                {
                    'key': column_definition.key,
                    'type': column_definition.type,
                    'name': column_definition.title,
                    'content': value,
                }
            )

        row = Row(cells=prepared_data)
        self._rows.append(row)
        log.debug('Row: %s was added to Table', row)

    def save(self, force: bool = True) -> None:
        """Save all changes in the table

        Args:
            force: Force to update properties without digest check.

        Returns:

        """
        super().save(force)

        row_requests: List[ChangeRowRequest] = []
        for row in self._rows:
            row_request = row.get_change_request()
            if row_request:
                row_requests.append(row_request)

        if not row_requests:
            return

        request = ChangeTableDataRequest(data=row_requests)
        api = SignalsNotebookApi.get_default_api()

        api.call(
            method='PATCH',
            path=(self._get_adt_endpoint(), self.eid),
            params={
                'digest': None if force else self.digest,
                'force': json.dumps(force),
            },
            data=request.json(exclude_none=True, by_alias=True),
        )

        self._reload_data()

    def get(self, value: Union[str, UUID], default: Any = None) -> Union[Row, Any]:
        """Get Row

        Args:
            value: name of id of Row
            default: default value if it doens't exist

        Returns:
            Union[Row, Any]
        """
        try:
            return self[value]
        except KeyError:
            log.debug('KeyError were caught. Default value returned')
            return default

    def get_content(self, content_type: str = ContentType.CSV.value) -> File:
        """Get Table content

        Args:
            content_type: Export resource format

        Returns:

        """
        self.ContentType(content_type)
        if content_type == self.ContentType.JSON.value:
            rows = []
            for item in self:
                row = {}
                for cell in item:
                    row[cell.name] = cell.content.dict()
                rows.append(row)
            return File(
                name=f'{self.name}.json',
                content=json.dumps({'data': rows}, default=str).encode('utf-8'),
                content_type=content_type,
            )

        api = SignalsNotebookApi.get_default_api()
        log.debug('Get content for: %s| %s', self.__class__.__name__, self.eid)

        response = api.call(
            method='GET',
            path=(self._get_endpoint(), self.eid, 'export'),
            params={
                'format': None,
            },
        )

        content_disposition = response.headers.get('content-disposition', '')
        _, params = cgi.parse_header(content_disposition)

        return File(
            name=params['filename'],
            content=response.content,
            content_type=response.headers.get('content-type'),
        )

    @classmethod
    def create(
        cls,
        *,
        container: Container,
        name: str,
        content: Optional[List[Dict[str, CellContentDict]]] = None,
        template: Optional[str] = None,
        digest: Optional[str] = None,
        force: bool = True,
    ) -> Entity:
        """Create Table Entity

        Args:
            container: Container where create new Table
            name: file name
            content: Table content
            template: template for table creation
            digest: Indicate digest of entity. It is used to avoid conflict while concurrent editing.
            force: Force to post attachment

        Returns:
            Table
        """
        log.debug('Create Table: %s...', cls.__name__)
        if template:
            api = SignalsNotebookApi.get_default_api()
            request = {
                'data': {
                    'type': EntityType.GRID,
                    'attributes': {'name': name},
                    'relationships': {
                        'ancestors': {'data': [{'type': EntityType.EXPERIMENT, 'id': container.eid}]},
                        'template': {'data': {'type': EntityType.GRID, 'id': template}},
                    },
                }
            }

            response = api.call(
                method='POST',
                path=(cls._get_endpoint(),),
                params={
                    'digest': digest,
                    'force': json.dumps(force),
                },
                json=request,
            )
            result = TableResponse(**response.json())
            table = cast(ResponseData, result.data).body
            log.debug('Entity: %s was created.', cls.__name__)
            if content:
                for row in content:
                    table.add_row(row)
                table.save()
            return table

        log.debug('There is no needful template. Table will be uploaded as *.csv File...')
        log.debug('Create table: %s with name: %s in Container: %s', cls.__name__, name, container.eid)
        return container.add_child(
            name=name,
            content=json.dumps({'data': content}, default=str).encode('utf-8'),
            content_type=cls.ContentType.JSON,
            force=force,
        )

    def get_html(self) -> str:
        """Get in HTML format

        Returns:
            Rendered template as a string
        """

        rows = []
        column_definitions = self.get_column_definitions_list()

        table_head = []
        for column_definition in column_definitions:
            table_head.append(column_definition.title)

        for row in self:
            reformatted_row = {}

            for column_definition in column_definitions:
                cell = cast(Row, row).get(column_definition.key, None)
                cell = cast(Cell, cell)
                reformatted_row[column_definition.title] = '' if cell is None else (cell.display or cell.value)

            rows.append(reformatted_row)

        template = env.get_template(self._template_name)
        log.info('Html template for %s:%s has been rendered.', self.__class__.__name__, self.eid)

        return template.render(name=self.name, table_head=table_head, rows=rows)

    def dump(self, base_path: str, fs_handler: FSHandler, alias: Optional[List[str]] = None) -> None:
        """Dump Table entity

        Args:
            base_path: content path where create dump
            fs_handler: FSHandler
            alias: Backup alias

        Returns:

        """
        log.debug('Dumping table: %s with name: %s...', self.eid, self.name)

        content = self.get_content(content_type=self.ContentType.JSON)
        column_definitions = self.get_column_definitions_list()

        metadata = {
            'file_name': content.name,
            'content_type': content.content_type,
            'columns': [item.title for item in column_definitions],
            **{k: v for k, v in self.dict().items() if k in ('name', 'description', 'eid')},
        }
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, 'metadata.json'),
            json.dumps(metadata, default=str),
            base_alias=alias + [self.name, '__Metadata'] if alias else None,
        )
        file_name = content.name
        data = content.content
        fs_handler.write(
            fs_handler.join_path(base_path, self.eid, file_name),
            data,
            base_alias=alias + [self.name, file_name] if alias else None,
        )
        log.debug('Table: %s was dumped successfully', self.eid, self.name)

    @classmethod
    def load(cls, path: str, fs_handler: FSHandler, parent: Container) -> None:
        """Load Table entity

        Args:
            path: content path
            fs_handler: FSHandler
            parent: Container where load Table entity

        Returns:

        """
        cls._load(path, fs_handler, parent)

    @classmethod
    def _load(cls, path: str, fs_handler: FSHandler, parent: Any) -> None:
        log.debug('Loading table from dump...')
        metadata_path = fs_handler.join_path(path, 'metadata.json')
        metadata = json.loads(fs_handler.read(metadata_path))
        content_path = fs_handler.join_path(path, metadata['file_name'])
        content_bytes = fs_handler.read(content_path)
        content = json.loads(content_bytes)
        rows = content['data']
        column_definitions = metadata.get('columns')
        templates = EntityStore.get_list(
            include_types=[EntityType.GRID], include_options=[EntityStore.IncludeOptions.TEMPLATE]
        )

        file_creation = True
        for item in templates:
            template = cast('Table', item)
            template_column_definitions = template.get_column_definitions_list()
            template_columns = [item.title for item in template_column_definitions]
            if set(template_columns) == set(column_definitions):
                file_creation = False
                cls.create(
                    container=parent,
                    name=metadata['name'],
                    template=template.eid,
                    content=rows,
                )
                break

        if file_creation:
            cls.create(container=parent, name=metadata['name'], content=rows, force=True)
        log.debug('Table was loaded to Container: %s', parent.eid)

    @classmethod
    def dump_templates(cls, base_path: str, fs_handler: FSHandler) -> None:
        """Dump Table templates

        Args:
            base_path: content path where create templates dump
            fs_handler: FSHandler

        Returns:

        """
        from signals_notebook.entities import EntityStore

        entity_type = cls._get_entity_type()

        templates = EntityStore.get_list(
            include_types=[entity_type], include_options=[EntityStore.IncludeOptions.TEMPLATE]
        )

        try:
            for item in templates:
                template = cast('Table', item)
                content = template.get_content(content_type=cls.ContentType.JSON)
                column_definitions = template.get_column_definitions_list()
                metadata = {
                    'file_name': content.name,
                    'content_type': content.content_type,
                    'columns': [item.title for item in column_definitions],
                    **{k: v for k, v in template.dict().items() if k in ('name', 'description', 'eid')},
                }
                fs_handler.write(
                    fs_handler.join_path(base_path, 'templates', entity_type, f'metadata_{template.name}.json'),
                    json.dumps(metadata),
                    ('Templates', entity_type.value, template.name),
                )
        except TypeError:
            pass


class TableResponse(Response[Table]):
    pass



from typing import Any, Literal

from pydantic import Field

from signals_notebook.common_types import MaterialType
from signals_notebook.materials.field import FieldContainer
from signals_notebook.materials.material import Material


class Batch(Material):
    type: Literal[MaterialType.BATCH] = Field(allow_mutation=False, default=MaterialType.BATCH)

    def __init__(self, **data: Any):
        fields = data.pop('fields', {})
        super().__init__(**data)

        self._material_fields = FieldContainer(self, self.library.batch_config.fields, **fields)
